import asyncio

from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, check_direct, get_db_path, \
    get_filename_and_ext
from ..base_functions.data_classes import DomainItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class ShareXCrawler:
    def __init__(self, *, include_id=False, quiet: bool, SQL_Helper: SQLHelper):
        self.include_id = include_id
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for ShareX scraper"""
        domain_obj = DomainItem(url.host.lower(), {})

        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)

        if await check_direct(url):
            url = url.with_name(url.name.replace('.md.', '.').replace('.th.', '.'))
            url_path = await get_db_path(url)
            complete = await self.SQL_Helper.check_complete_singular("anonfiles", url_path)
            filename, ext = await get_filename_and_ext(url.name)
            media_item = MediaItem(url, url, complete, filename, ext)
            await domain_obj.add_media("Loose ShareX Files", media_item)
        elif "album" in url.parts or "a" in url.parts:
            await self.parse(session=session, url=url, domain_obj=domain_obj)
        elif "albums" in url.parts:
            await self.get_albums(session, url, domain_obj)
        elif 'image' in url.parts or 'img' in url.parts or 'images' in url.parts:
            await self.get_singular(session, url, domain_obj)
        else:
            await self.parse_profile(session, url, domain_obj)

        url_path = await get_db_path(url)
        await self.SQL_Helper.insert_domain("sharex", url_path, domain_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return domain_obj

    async def get_albums(self, session: ScrapeSession, url: URL, domain_obj: DomainItem):
        """Handles scraping for Albums"""
        try:
            soup = await session.get_BS4(url)
            albums = soup.select("a[class='image-container --media']")
            for album in albums:
                album_url = URL(album.get('href'))
                await self.parse(session=session, url=album_url, domain_obj=domain_obj)

            next_page = soup.select_one('li.pagination-next a')
            if not next_page:
                next_page = soup.select_one('a[data-pagination=next]')
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    next_page = URL(next_page)
                    await self.get_albums(session, next_page, domain_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_singular(self, session: ScrapeSession, url: URL, domain_obj: DomainItem):
        """Handles scraping for singular files"""
        await asyncio.sleep(1)
        try:
            soup = await session.get_BS4(url)
            link = URL(soup.select_one("input[id=embed-code-2]").get('value'))
            link = link.with_name(link.name.replace('.md.', '.').replace('.th.', '.'))

            url_path = await get_db_path(link)
            complete = await self.SQL_Helper.check_complete_singular("sharex", url_path)
            filename, ext = await get_filename_and_ext(link.name)
            media_item = MediaItem(link, url, complete, filename, ext)
            await domain_obj.add_media("Loose ShareX Files", media_item)
        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_sub_album_links(self, session: ScrapeSession, url: URL, og_title: str, domain_obj: DomainItem):
        try:
            soup = await session.get_BS4(url)
            albums = soup.select("div[class=pad-content-listing] div")
            for album in albums:
                album_url = album.get('data-url-short')
                if album_url is not None:
                    album_url = URL(album_url)
                    await self.parse(session=session, url=album_url, og_title=og_title, domain_obj=domain_obj)
        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def parse_profile(self, session: ScrapeSession, url: URL, domain_obj: DomainItem):
        """Handles scraping for profiles"""
        try:
            soup = await session.get_BS4(url)
            title = soup.select_one("div[class=header] h1 strong").get_text()
            if title is None:
                title = url.name
            elif self.include_id:
                titlep2 = url.name
                title = title + " - " + titlep2
            title = await make_title_safe(title.replace(r"\n", "").strip())
            await self.get_list_links(session, url, title, domain_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_list_links(self, session: ScrapeSession, url: URL, title: str, domain_obj: DomainItem):
        """Gets final links and adds to domain_obj"""
        try:
            soup = await session.get_BS4(url)
            if url.host == 'jpg.fish' or url.host == 'jpg.church':
                links = soup.select("a[href*=img] img")
            else:
                links = soup.select("a[href*=image] img")
            for link in links:
                link = URL(link.get('src'))
                link = link.with_name(link.name.replace('.md.', '.').replace('.th.', '.'))

                try:
                    filename, ext = await get_filename_and_ext(link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(link))
                    continue
                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("sharex", url_path)
                media_item = MediaItem(link, url, complete, filename, ext)
                await domain_obj.add_media(title, media_item)

            next_page = soup.select_one('li.pagination-next a')
            if not next_page:
                next_page = soup.select_one('a[data-pagination=next]')
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    next_page = URL(next_page)
                    await self.get_list_links(session, next_page, title, domain_obj)
        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def parse(self, *, session: ScrapeSession, url: URL, og_title=None, domain_obj: DomainItem):
        try:
            soup = await session.get_BS4(url)

            title = soup.select_one("a[data-text=album-name]").get_text()
            if title is None:
                title = url.name
            elif self.include_id:
                titlep2 = url.name
                title = title + " - " + titlep2
            title = await make_title_safe(title.replace(r"\n", "").strip())

            if og_title is not None:
                title = og_title + "/" + title

            try:
                sub_albums = URL(soup.select_one("a[id=tab-sub-link]").get("href"))
                await self.get_sub_album_links(session, sub_albums, title, domain_obj)
            finally:
                list_recent = URL(soup.select_one("a[id=list-most-recent-link]").get("href"))
                await self.get_list_links(session, list_recent, title, domain_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class HGameCGCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Basic director for HGameCG"""
        album_obj = AlbumItem("Loose HGamesCG Files", [])

        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        await self.get_album(session, url, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        url_path = await get_db_path(url)
        await self.SQL_Helper.insert_album("hgamecg", url_path, album_obj)
        return album_obj

    async def get_album(self, session: ScrapeSession, url: URL, album_obj: AlbumItem):
        """Handles album scraping, adds media items to the album_obj"""
        try:
            soup = await session.get_BS4(url)
            title = await make_title_safe(soup.select_one("div[class=navbar] h1").get_text())
            await album_obj.set_new_title(title)

            images = soup.select("div[class=image] a")
            for image in images:
                image = image.get('href')
                image = URL("https://" + url.host + image)
                link = await self.get_image(session, image)

                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("hgamecg", url_path)

                filename, ext = await get_filename_and_ext(link.name)
                media_item = MediaItem(link, image, complete, filename, ext)
                await album_obj.add_media(media_item)

            next_page = soup.find("a", text="Next Page")
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    next_page = URL("https://" + url.host + next_page)
                    await self.get_album(session, next_page, album_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_image(self, session: ScrapeSession, url: URL):
        """Gets image link from the given url."""
        try:
            soup = await session.get_BS4(url)
            image = soup.select_one("div[class=hgamecgimage] img")
            image = URL(image.get('src'))
            return image
        except Exception as e:
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)



import asyncio
import os
from pathlib import Path
from .base_functions import FILE_FORMATS, log


class Sorter:
    def __init__(self, download_dir: Path, sorted_downloads: Path, audio_dir: str, image_dir: str, video_dir: str,
                 other_dir: str):
        self.download_dir = download_dir
        self.sorted_downloads = sorted_downloads

        self.audio_dir = audio_dir
        self.image_dir = image_dir
        self.video_dir = video_dir
        self.other_dir = other_dir

        self.audio = 0
        self.images = 0
        self.videos = 0
        self.other = 0

    async def find_files_in_dir(self, directory: Path):
        file_list = []
        for x in directory.iterdir():
            if x.is_file():
                file_list.append(x)
            elif x.is_dir():
                file_list.extend(await self.find_files_in_dir(x))
        return file_list

    async def sort(self):
        for folder in self.download_dir.iterdir():
            if not folder.is_dir():
                continue

            audio_destination = Path(self.audio_dir.format(sort_dir=self.sorted_downloads, base_dir=folder.name))
            image_destination = Path(self.image_dir.format(sort_dir=self.sorted_downloads, base_dir=folder.name))
            video_destination = Path(self.video_dir.format(sort_dir=self.sorted_downloads, base_dir=folder.name))
            other_destination = Path(self.other_dir.format(sort_dir=self.sorted_downloads, base_dir=folder.name))

            files = await self.find_files_in_dir(folder)
            for file in files:
                ext = file.suffix.lower()
                if ext == '.part':
                    continue
                elif ext in FILE_FORMATS['Audio']:
                    audio_destination.mkdir(parents=True, exist_ok=True)
                    await self.move_cd(file, audio_destination)
                    self.audio += 1
                elif ext in FILE_FORMATS['Images']:
                    image_destination.mkdir(parents=True, exist_ok=True)
                    await self.move_cd(file, image_destination)
                    self.images += 1
                elif ext in FILE_FORMATS['Videos']:
                    video_destination.mkdir(parents=True, exist_ok=True)
                    await self.move_cd(file, video_destination)
                    self.videos += 1
                else:
                    other_destination.mkdir(parents=True, exist_ok=True)
                    await self.move_cd(file, other_destination)
                    self.other += 1
        await asyncio.sleep(5)
        await self.purge_dir(str(self.download_dir))
        await log(f"[green]Organized: {self.audio} Audio Files[/green]", quiet=False)
        await log(f"[green]Organized: {self.images} Image Files[/green]", quiet=False)
        await log(f"[green]Organized: {self.videos} Video Files[/green]", quiet=False)
        await log(f"[green]Organized: {self.other} Other Files[/green]", quiet=False)

    async def move_cd(self, file: Path, dest: Path):
        try:
            file.rename((dest / file.name))
        except FileExistsError:
            stem = file.stem
            temp_stem = stem
            ext = file.suffix
            i = 1
            while (dest / f"{temp_stem}{ext}").is_file() and file.is_file():
                if file.stat().st_size == (dest / file.name).stat().st_size:
                    file.unlink()
                temp_stem = f"{stem} ({i}){ext}"
                i += 1
            if file.is_file():
                file.rename((dest / f"{temp_stem}{ext}"))

    async def purge_dir(self, directory: str):
        dir_tree = list(os.walk(directory, topdown=False))
        for tree_element in dir_tree:
            sub_dir = tree_element[0]
            is_empty = not len(os.listdir(sub_dir))
            if is_empty:
                os.rmdir(sub_dir)



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class EHentaiCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for E-Hentai"""
        album_obj = AlbumItem("Loose EHentai Files", [])

        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        if "g" in url.parts:
            await self.get_album(session, url, album_obj)
        elif "s" in url.parts:
            await self.get_image(session, url, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        url_path = await get_db_path(url)
        await self.SQL_Helper.insert_album("e-hentai", url_path, album_obj)
        return album_obj

    async def get_album(self, session: ScrapeSession, url: URL, album_obj: AlbumItem):
        """Gets links from an album"""
        try:
            soup = await session.get_BS4(url)
            title = await make_title_safe(soup.select_one("h1[id=gn]").get_text())
            await album_obj.set_new_title(title)

            images = soup.select("div[class=gdtm] div a")
            for image in images:
                image = image.get('href')
                image = URL(image)
                await self.get_image(session, image, album_obj)

            next_page_opts = soup.select('td[onclick="document.location=this.firstChild.href"]')
            next_page = None
            for maybe_next in next_page_opts:
                if maybe_next.get_text() == ">":
                    next_page = maybe_next.select_one('a')
                    break
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    await self.get_album(session, next_page, album_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_image(self, session: ScrapeSession, url: URL, album_obj: AlbumItem):
        """Gets media items from image links"""
        try:
            soup = await session.get_BS4(url)
            image = soup.select_one("img[id=img]")
            link = URL(image.get('src'))

            url_path = await get_db_path(link, 'e-hentai')
            complete = await self.SQL_Helper.check_complete_singular("e-hentai", url_path)

            filename, ext = await get_filename_and_ext(link.name)
            media_item = MediaItem(link, image, complete, filename, ext)
            await album_obj.add_media(media_item)
        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)




from yarl import URL

from ..base_functions.base_functions import log, logger, check_direct, get_filename_and_ext, get_db_path, \
    make_title_safe
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class ImgBoxCrawler:
    def __init__(self, *, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director func for ImgBox scraping"""
        album_obj = AlbumItem("Loose ImgBox Files", [])
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)

        try:
            if await check_direct(url):
                filename, ext = await get_filename_and_ext(url.name)
                url_path = await get_db_path(url)
                complete = await self.SQL_Helper.check_complete_singular("imgbox", url_path)
                media_item = MediaItem(url, url, complete, filename, ext)
                await album_obj.add_media(media_item)

            elif "g" in url.parts:
                title, images = await self.folder(session, url)
                if not title:
                    title = url.raw_name
                title = await make_title_safe(title)
                await album_obj.set_new_title(title)
                for img in images:
                    try:
                        filename, ext = await get_filename_and_ext(img.name)
                    except NoExtensionFailure:
                        logger.debug("Couldn't get extension for %s", str(img))
                        continue
                    url_path = await get_db_path(img)
                    complete = await self.SQL_Helper.check_complete_singular("imgbox", url_path)
                    media_item = MediaItem(img, url, complete, filename, ext)
                    await album_obj.add_media(media_item)
            else:
                img = await self.singular(session, url)
                filename, ext = await get_filename_and_ext(img.name)
                url_path = await get_db_path(img)
                complete = await self.SQL_Helper.check_complete_singular("imgbox", url_path)
                media_item = MediaItem(img, url, complete, filename, ext)
                await album_obj.add_media(media_item)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

        await self.SQL_Helper.insert_album("imgbox", url.path, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj

    async def folder(self, session: ScrapeSession, url: URL):
        """Gets links from a folder"""
        soup = await session.get_BS4(url)
        output = []
        title = soup.select_one("div[id=gallery-view] h1").get_text()

        images = soup.find('div', attrs={'id': 'gallery-view-content'})
        images = images.findAll("img")
        for link in images:
            link = link.get('src').replace("thumbs", "images").replace("_b", "_o")
            output.append(URL(link))

        return title, output

    async def singular(self, session: ScrapeSession, url: URL):
        """Gets individual links"""
        soup = await session.get_BS4(url)
        link = URL(soup.select_one("img[id=img]").get('src'))
        return link



import asyncio
import collections
import time

from datetime import datetime
from types import TracebackType
from typing import Optional, Callable, Awaitable, Any, Type


class AsyncRateLimiter:
    """Provides rate limiting for an operation with a configurable number of requests for a time period."""

    __lock: asyncio.Lock
    callback: Optional[Callable[[float], Awaitable[Any]]]
    max_calls: int
    period: float
    calls: collections.deque

    def __init__(
        self,
        max_calls: int,
        period: float = 1.0,
        callback: Optional[Callable[[float], Awaitable[Any]]] = None,
    ):
        if period <= 0:
            raise ValueError("Rate limiting period should be > 0")
        if max_calls <= 0:
            raise ValueError("Rate limiting number of calls should be > 0")
        self.calls = collections.deque()

        self.period = period
        self.max_calls = max_calls
        self.callback = callback
        self.__lock = asyncio.Lock()

    async def __aenter__(self) -> "AsyncRateLimiter":
        async with self.__lock:
            if len(self.calls) >= self.max_calls:
                until = datetime.utcnow().timestamp() + self.period - self._timespan
                if self.callback:
                    asyncio.ensure_future(self.callback(until))
                sleep_time = until - datetime.utcnow().timestamp()
                if sleep_time > 0:
                    await asyncio.sleep(sleep_time)
            return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        async with self.__lock:
            # Store the last operation timestamp.
            self.calls.append(datetime.utcnow().timestamp())

            while self._timespan >= self.period:
                self.calls.popleft()

    @property
    def _timespan(self) -> float:
        return self.calls[-1] - self.calls[0]


async def throttle(self, delay: int, host: str) -> None:
    """Throttles requests to domains by a parameter amount of time"""
    if delay is None or delay == 0:
        return

    key: Optional[str] = None
    while True:
        if key is None:
            key = f'throttle:{host}'
        now = time.time()
        last = self.throttle_times.get(key, 0.0)
        elapsed = now - last

        if elapsed >= delay:
            self.throttle_times[key] = now
            return

        remaining = delay - elapsed + 0.1
        await asyncio.sleep(remaining)



import http
from typing import Union

from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import DomainItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class GoFileCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

        self.api_address = URL("https://api.gofile.io")
        self.token = None

    async def get_token(self, session: ScrapeSession):
        """Creates an anon gofile account to use."""
        if self.token:
            return

        try:
            json_obj = await session.get_json(self.api_address / "createAccount")
            if json_obj["status"] == "ok":
                self.token = json_obj["data"]["token"]
                await self.set_cookie(session)
            else:
                raise
        except Exception as e:
            logger.debug("Error encountered while getting GoFile token", exc_info=True)
            await log(f"[red]Error: Couldn't generate GoFile token[/red]", quiet=self.quiet)
            logger.debug(e)

    async def set_cookie(self, session: ScrapeSession):
        """Sets the given token as a cookie into the session (and client)"""
        client_token = self.token
        morsel = http.cookies.Morsel()
        morsel['domain'] = 'gofile.io'
        morsel.set('accountToken', client_token, client_token)
        session.client_session.cookie_jar.update_cookies({'gofile.io': morsel})

    async def fetch(self, session: ScrapeSession, url: URL):
        """Basic director for actual scraping"""
        try:
            await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
            domain_obj = DomainItem("gofile", {})
            content_id = url.name
            results = await self.get_links(session, url, content_id, None)
            for title, media_item in results:
                await domain_obj.add_media(title, media_item)

            url_path = await get_db_path(url)
            await self.SQL_Helper.insert_domain("gofile", url_path, domain_obj)
            await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
            return domain_obj

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_links(self, session: ScrapeSession, url: URL, content_id: str, title=None):
        """Gets links from the given url, creates media_items"""
        results = []
        params = {
            "token": self.token,
            "contentId": content_id,
            "websiteToken": "12345",
        }
        content = await session.get_json_with_params(self.api_address / "getContent", params)
        if content["status"] != "ok":
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            return results

        content = content['data']
        if title:
            title = title + "/" + await make_title_safe(content["name"])
        else:
            title = await make_title_safe(content["name"])

        contents: dict[str, dict[str, Union[str, int]]] = content["contents"]
        sub_folders = []
        for val in contents.values():
            if val["type"] == "folder":
                sub_folders.append(val['code'])
            else:
                link = URL(val["link"]) if val["link"] != "overloaded" else URL(val["directLink"])
                try:
                    filename, ext = await get_filename_and_ext(val['name'])
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(link))
                    continue
                link_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("gofile", link_path)
                media_item = MediaItem(link, url, complete, filename, ext)
                results.append([title, media_item])
        for sub_folder in sub_folders:
            results.extend(await self.get_links(session, url, sub_folder, title))
        return results



from __future__ import annotations

import asyncio
import json
import logging
import ssl
from pathlib import Path

import aiofiles
from bs4 import BeautifulSoup
from rich.progress import TaskID
from yarl import URL

import aiohttp
import certifi

from cyberdrop_dl.downloader.progress_definitions import file_progress
from .rate_limiting import AsyncRateLimiter, throttle
from ..base_functions.base_functions import logger
from ..base_functions.data_classes import FileLock, MediaItem
from ..base_functions.error_classes import DownloadFailure, InvalidContentTypeFailure


class Client:
    """Creates a 'client' that can be referenced by scraping or download sessions"""
    def __init__(self, ratelimit: int, throttle: int, secure: bool, connect_timeout: int):
        self.connect_timeout = connect_timeout
        self.ratelimit = ratelimit
        self.throttle = throttle
        self.simultaneous_session_limit = asyncio.Semaphore(50)
        self.user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/109.0'
        self.verify_ssl = secure
        self.ssl_context = ssl.create_default_context(cafile=certifi.where()) if secure else False
        self.cookies = aiohttp.CookieJar(quote_cookie=False)


class ScrapeSession:
    """AIOHTTP operations for scraping"""
    def __init__(self, client: Client) -> None:
        self.client = client
        self.rate_limiter = AsyncRateLimiter(self.client.ratelimit)
        self.headers = {"user-agent": client.user_agent}
        self.timeouts = aiohttp.ClientTimeout(total=5 * 60, connect=self.client.connect_timeout)
        self.client_session = aiohttp.ClientSession(headers=self.headers, raise_for_status=True, cookie_jar=self.client.cookies, timeout=self.timeouts)

    async def get_BS4(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    content_type = response.headers.get('Content-Type')
                    if 'text' in content_type.lower() or 'html' in content_type.lower():
                        text = await response.text()
                        soup = BeautifulSoup(text, 'html.parser')
                        return soup
                    else:
                        raise InvalidContentTypeFailure(message=f"Received {content_type}, was expecting text")

    async def get_BS4_and_url(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    text = await response.text()
                    soup = BeautifulSoup(text, 'html.parser')
                    return soup, URL(response.url)

    async def get_json(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    content = json.loads(await response.content.read())
                    return content

    async def get_json_with_params(self, url: URL, params: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context, params=params) as response:
                    content = json.loads(await response.content.read())
                    return content

    async def get_text(self, url: URL):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, ssl=self.client.ssl_context) as response:
                    text = await response.text()
                    return text

    async def post(self, url: URL, data: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.post(url, data=data, headers=self.headers, ssl=self.client.ssl_context) as response:
                    content = json.loads(await response.content.read())
                    return content

    async def get_no_resp(self, url: URL, headers: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context) as response:
                    pass

    async def post_data_no_resp(self, url: URL, data: dict):
        async with self.client.simultaneous_session_limit:
            async with self.rate_limiter:
                async with self.client_session.post(url, data=data, headers=self.headers, ssl=self.client.ssl_context) as response:
                    pass

    async def exit_handler(self):
        try:
            await self.client_session.close()
        except Exception as e:
            logging.debug(f"Failed to close session.")


class DownloadSession:
    """AIOHTTP operations for downloading"""
    def __init__(self, client: Client):
        self.client = client
        self.headers = {"user-agent": client.user_agent}
        self.timeouts = aiohttp.ClientTimeout(total=5 * 60, connect=self.client.connect_timeout)
        self.client_session = aiohttp.ClientSession(headers=self.headers, raise_for_status=True, cookie_jar=self.client.cookies, timeout=self.timeouts)
        self.throttle_times = {}

    async def download_file(self, media: MediaItem, file: Path, current_throttle: int, resume_point: int,
                            File_Lock: FileLock, proxy: str, headers: dict, original_filename: str,
                            file_task: TaskID):
        headers['Referer'] = str(media.referer)
        headers['user-agent'] = self.client.user_agent
        await throttle(self, current_throttle, media.url.host)
        async with self.client_session.get(media.url, headers=headers, ssl=self.client.ssl_context,
                                           raise_for_status=True, proxy=proxy) as resp:
            content_type = resp.headers.get('Content-Type')
            if 'text' in content_type.lower() or 'html' in content_type.lower():
                logger.debug("Server for %s is experiencing issues, you are being ratelimited, or cookies have expired", str(media.url))
                await File_Lock.remove_lock(original_filename)
                raise DownloadFailure(code=resp.status, message="Unexpectedly got text as response")

            total = int(resp.headers.get('Content-Length', str(0))) + resume_point
            file.parent.mkdir(parents=True, exist_ok=True)

            file_progress.update(file_task, total=total)
            file_progress.advance(file_task, resume_point)

            async with aiofiles.open(file, mode='ab') as f:
                async for chunk, _ in resp.content.iter_chunks():
                    await asyncio.sleep(0)
                    await f.write(chunk)
                    file_progress.advance(file_task, len(chunk))

    async def get_filesize(self, url: URL, referer: str, current_throttle: int):
        headers = {'Referer': referer, 'user-agent': self.client.user_agent}
        await throttle(self, current_throttle, url.host)
        async with self.client_session.get(url, headers=headers, ssl=self.client.ssl_context,
                                           raise_for_status=True) as resp:
            total_size = int(resp.headers.get('Content-Length', str(0)))
            return total_size



from yarl import URL

from ..base_functions.base_functions import log, logger, get_db_path, get_filename_and_ext
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class SaintCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Basic director for saint scraping"""
        album_obj = AlbumItem("Loose Saint Files", [])
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)

        try:
            soup = await session.get_BS4(url)
            link = URL(soup.select_one('video[id=main-video] source').get('src'))
            url_path = await get_db_path(link)
            complete = await self.SQL_Helper.check_complete_singular("saint", url_path)
            filename, ext = await get_filename_and_ext(link.name)
            media_item = MediaItem(link, url, complete, filename, ext)
            await album_obj.add_media(media_item)
            await self.SQL_Helper.insert_album("saint", url_path, album_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj
